/* simple R compiler */

#define USE_RINTERNALS
#include <R.h>
#include <Rinternals.h>

#ifndef NATFSXP
#warning "R is lacking NATFSXP support, falling back to .Call interface (untested)"
#define CALL_SIG
#endif

/* ---- debug tools ---- */

static inline void print_exp(SEXP ex) { /* print(quote(`ex`)) */
  eval(LCONS(install("print"),CONS(LCONS(install("quote"),CONS(ex, R_NilValue)),R_NilValue)),R_GlobalEnv);
}

static inline void print_val(SEXP v) { /* print(`v`) */
  eval(LCONS(install("print"),CONS(v, R_NilValue)), R_GlobalEnv);
}

/* ---- language info ---- */

#define FNT_MATH1   1
#define FNT_ARITHOP 2
#define FNT_SPECIAL 3

#define FNT_IMPURE  0x10000

#define FNT_IS_SPECIAL(X) (((X)&0xff)==FNT_SPECIAL)
#define FNT_IS_PURE(X) (((X)&FNT_IMPURE)==0)
#define FNT_CLASS(X) ((X)&0xff)

/* some special functions that are optimized */
#define FNT_COMPOUND (FNT_SPECIAL | 0x0100)
#define FNT_FOR      (FNT_SPECIAL | 0x0200)
#define FNT_COLON    (FNT_SPECIAL | 0x0300)
#define FNT_COLLECT  (FNT_SPECIAL | 0x0400)
#define FNT_ASSIGN   (FNT_SPECIAL | 0x0500)

typedef struct fn_info_s {
  const char *fn;
  int type;
} fn_info_t;

#define INT_FNS 39

static const fn_info_t int_funs[INT_FNS] = {
  { "{",        FNT_COMPOUND}, { "for",      FNT_FOR     }, { ":",          FNT_COLON   },

  { "=",        FNT_ASSIGN  }, { "<-",       FNT_ASSIGN  },

  { "+",        FNT_ARITHOP }, { "-",        FNT_ARITHOP }, { "*",          FNT_ARITHOP },
  { "/",        FNT_ARITHOP }, { "^",        FNT_ARITHOP }, { "%%",         FNT_ARITHOP },

  { "(",        FNT_COLLECT },

  { "floor",    FNT_MATH1   }, { "ceil",     FNT_MATH1   }, { "sqrt",       FNT_MATH1   }, 
  { "sign",     FNT_MATH1   }, { "trunc",    FNT_MATH1   }, { "exp",        FNT_MATH1   }, 
  { "expm1",    FNT_MATH1   }, { "log1p",    FNT_MATH1   }, { "cos",        FNT_MATH1   }, 
  { "sin",      FNT_MATH1   }, { "tan",      FNT_MATH1   }, { "acos",       FNT_MATH1   }, 
  { "asin",     FNT_MATH1   }, { "cosh",     FNT_MATH1   }, { "sinh",       FNT_MATH1   },
  { "tanh",     FNT_MATH1   }, { "acosh",    FNT_MATH1   }, { "asinh",      FNT_MATH1   }, 
  { "atanh",    FNT_MATH1   }, { "lgammafn", FNT_MATH1   }, { "gammafn",    FNT_MATH1   }, 
  { "digamma",  FNT_MATH1   }, { "trigamma", FNT_MATH1   }, { "gamma_cody", FNT_MATH1   },

  { "print",    FNT_IMPURE  }, { "plot",     FNT_IMPURE  }, { "show",       FNT_IMPURE  }
};

static inline int fn_id(const char *fn) {
  int i = 0;
  while (i < INT_FNS) {
    if (!strcmp(int_funs[i].fn, fn)) return i;
    i++;
  }
  return -1;
}

static inline int fn_type(const char *fn) {
  int i = fn_id(fn);
  return (i>=0)?int_funs[i].type:0;
}

/* ---- linked buffer implementation ---- */
#define LBUF_SIZE 8192
struct lbuf {
  char buf[LBUF_SIZE];
  int len;
  struct lbuf *next;
  struct lbuf *head;
};

static struct lbuf *insert_str(struct lbuf *lb, const char *s) {
  int i = strlen(s);
  while (i > 0) {
    int space = LBUF_SIZE - lb->len - 1;
    if (i < space) {
      strcpy(lb->buf + lb->len, s);
      lb->len += i;
      return lb;
    }
    if (space > 0) {
      memcpy(lb->buf + lb->len, s, space);
      i -= space;
      s += space;
      lb->buf[LBUF_SIZE-1] = 0;
      lb->len = LBUF_SIZE - 1;
    }
    {
      struct lbuf *pnext = lb->next;
      lb->next = (struct lbuf*) R_alloc(1, sizeof(struct lbuf));
      lb->next->head = lb->head;
      lb = lb->next;
      lb->len = 0;
      lb->next = pnext;
    }
  }
  return lb;
}

static struct lbuf *pre_str(struct lbuf *lb, const char *s) {
  struct lbuf *pre = (struct lbuf*) R_alloc(1, sizeof(struct lbuf));
  pre->len = 0;
  pre->next = lb;
  pre->head = pre;
  {
    struct lbuf *n = lb;
    while (n) { n->head = pre; n = n->next; }
  }
  return insert_str(pre, s);
}

static struct lbuf *add_str(struct lbuf *lb, const char *s) {
  while (lb->next) lb = lb->next;
  return insert_str(lb, s);
}

static char *concat_str(struct lbuf *lb) {
  long totl = 0;
  char *str, *s;
  struct lbuf *c = lb->head;
  while (c) {
    totl += c->len;
    c = c->next;
  }
  totl++;
  s = str = (char*) malloc(totl);
  c = lb->head;
  while (c) {
    memcpy(s, c->buf, c->len);
    s+=c->len;
    c = c->next;
  }
  *s = 0;
  return str;
}

/* --- symbol table chain --- */

struct stchain {
  SEXP symtab;
  struct stchain *next;
};

static struct stchain* new_stchain(SEXP symtab, struct stchain *next) {
  struct stchain* nc = (struct stchain*) R_alloc(1, sizeof(struct stchain));
  nc->symtab = symtab;
  nc->next = next;
  return nc;
}

static void free_symtab(SEXP t);
static SEXP find_sym(SEXP symtab, SEXP sym);

static struct stchain* free_stchain(struct stchain *stc) {
  struct stchain *n = stc?stc->next:0;
  if (stc->symtab && stc->symtab != R_NilValue)
    free_symtab(stc->symtab);
  /* if (stc) free(stc); */
  return n;
}  

static SEXP find_sym_rec(struct stchain *stc, SEXP sym) {
  while (stc) {
    if (stc->symtab != R_NilValue) {
      SEXP p = find_sym(stc->symtab, sym);
      if (p) return p;
    }
    stc = stc -> next;
  }
  return 0;
}

/* --- compier info structure --- */

struct compinfo {
  SEXP gltab; /* globals table */
  SEXP ctab;  /* constants table */
  int const_ptr; /* constants counter */
  struct stchain *lstc; /* locals symbol table chain */
  struct lbuf *outbuf;
  char usedfn[INT_FNS];
};

/* --- symbol table implementation --- */

#define STYPE(A) ((A)&255)
#define SFLAG(A) ((A)&0xff00)
#define SET_SFLAG(A,B) ((A)=(((A)^((A)&0xff00))|((B)&0xff00)))
#define SET_STYPE(A,B) ((A)=(((A)^((A)&255))|((B)&255)))
#define ADD_SFLAG(A,B) ((A)=((A)|((B)&0xffff00)))

#define STY_NUM    0x0001
#define STY_INT    (0x0000|STY_NUM)
#define STY_REAL   (0x0002|STY_NUM)
#define STY_LGL    0x0004
#define STY_CPLX   0x0008
#define STY_CHAR   0x0010
#define STY_OTHER  0x0020  /* fixed type, but none of the above - see symbol value for exact type */
#define STY_NEW    0x0040  /* new local symbol - temporary only */

#define STY_UNKN   0x0000  /* unknown type - no assumptions can be made */

#define SFL_CONST  0x0100  /* constant */
#define SFL_SCAL   0x0200  /* scalar (i.e. exactly length 1) */
#define SFL_UNB    0x0400  /* unbound, i.e. have to use `get' */
#define SFL_ATT    0x0800  /* has attribute(s) */
#define SFL_GCONST 0x1000  /* global constant (in constants table) */
#define SFL_USED   0x2000  /* this symbol was referenced at least once */

/* value, flags */
#define S2S(A) (INTEGER(CADR(A))[0])
#define SSTYPE(A) STYPE(S2S(A))
#define SSFLAG(A) SFLAG(S2S(A))
#define SET_SSFLAG(A,B) SET_SFLAG(S2S(A),B)
#define ADD_SSFLAG(A,B) ADD_SFLAG(S2S(A),B)
#define SET_SSTYPE(A,B) SET_SFLAG(S2S(A),B)
#define SVAL(A) (CAR(A))
#define SET_SVAL(A, B) (SETCAR((A),(B)))
#define COPY_SSINFO(A, B) {SETCAR(A, duplicate(CAR(B))); SETCDR(A, duplicate(CDR(B))); if (TAG(B)!=R_NilValue) SET_TAG(A, duplicate(TAG(B)))}

static inline SEXP new_si(SEXP value, int flags) {
  SEXP f, c, r;
  PROTECT(f=ScalarInteger(flags));
  PROTECT(c=CONS(f, R_NilValue));
  r=CONS(value, c);
  UNPROTECT(2);
  return r;
}

static SEXP find_sym(SEXP symtab, SEXP sym) {
  SEXP c = CAR(symtab);
  /* Rprintf("find_sym(%08x) ", (int) symtab); */
  while (c != R_NilValue) {
    /* Rprintf("<%08x:%08x:%08x:%08x> ", (int) c, (int) CAR(c), (int) CDR(c), (int) TAG(c)); */
    if (TAG(c) == sym) return CAR(c);
    c = CDR(c);
  }
 return 0;
}

static SEXP add_sym(SEXP symtab, SEXP sym, SEXP info) {
  SEXP sh = CAR(symtab);
  PROTECT(sym);
  PROTECT(info);
  {
    SEXP ns = CONS(info, sh);
    PROTECT(ns);
    SET_TAG(ns, sym);
    SETCAR(symtab, ns);
    /* Rprintf("add_sym(%08x, sym=%08x)=%08x ", (int)*symtab, (int)sym, (int)ns); */
    UNPROTECT(3);
    return CAR(ns);
  }
}

static SEXP new_symtab() {
  SEXP t = CONS(R_NilValue, R_NilValue);
  R_PreserveObject(t);
  return t;
}

static void free_symtab(SEXP t) {
  R_ReleaseObject(t);
}


/* --- comp info handling --- */

static void initCI(struct compinfo *ci) {
  ci->gltab=new_symtab();
  ci->ctab=new_symtab();
  ci->outbuf = (struct lbuf*) R_alloc(1, sizeof(struct lbuf));
  ci->outbuf->head = ci->outbuf;
  ci->outbuf->len = 0;
  ci->outbuf->next = 0;
  ci->lstc = 0;
  ci->const_ptr = 0;
  memset(ci->usedfn, 0, INT_FNS);
}

static void freeCI(struct compinfo *ci) {
  free_symtab(ci->gltab);
}

static void comp_out(struct compinfo *ci, const char *fmt, ...) {
  char ib[1024];
  va_list(ap);
  va_start(ap, fmt);
  ib[1023]=0;
  vsnprintf(ib, 1023, fmt, ap);
  va_end(ap);
  if (*ib) ci->outbuf=add_str(ci->outbuf, ib);
}

/* --- actual compiler --- */

static void add_const(struct compinfo *ci, SEXP v, int len, char *res, SEXP *resy) {
  ci->const_ptr++;
  sprintf(res, "_const_%d", ci->const_ptr);
  resy[0]=add_sym(ci->ctab, install(res), new_si(v, STY_OTHER|SFL_CONST|SFL_GCONST));
}

static int classify_val(struct compinfo *ci, SEXP v, char *res, SEXP *resy) {
  int st = TYPEOF(v);
  int fl = (ATTRIB(v)==R_NilValue)?0:SFL_ATT;
  if (!res) return TRUE; /* if no result is needed, get out right away */
  *res = 0;
  resy[0]=0;
  switch (st) {
  case INTSXP:
    fl |= STY_INT;
    if (LENGTH(v)==1) {
      fl |= SFL_SCAL;
      sprintf(res, "ScalarInteger(%d)", INTEGER(v)[0]);
    } else add_const(ci, v, LENGTH(v)*sizeof(int), res, resy);
    break;
  case LGLSXP:
    fl |= STY_LGL;
    if (LENGTH(v)==1) {
      fl |= SFL_SCAL;
      sprintf(res, "ScalarLogical(%d)", INTEGER(v)[0]);
    } else add_const(ci, v, LENGTH(v)*sizeof(int), res, resy);
    break;
  case REALSXP:
    fl |= STY_REAL;
    if (LENGTH(v)==1) {
      fl |= SFL_SCAL;
      sprintf(res, "ScalarReal(%f)", REAL(v)[0]);
    } else add_const(ci, v, LENGTH(v)*sizeof(double), res, resy);
    break;
  case STRSXP:
    fl |= STY_CHAR;
    if (LENGTH(v)==1) {
      fl |= SFL_SCAL;
      sprintf(res, "mkString(\"%s\")", CHAR(STRING_ELT(v, 0)));      
    }
    break;
  default:
    fl |= STY_OTHER;
  }
  fl |= SFL_CONST; /* we're taking all global symbols as const for now */
  if (!resy[0])
    resy[0] = new_si(v, fl);
  return TRUE;
}

/* c_fn: compile a language construct
   e    = language construct (LANGSXP) to compile
   lp   = local ptr (incldeased when local variables are allocated)
   res  = result (as text) of this construct
   resy = result (as symbol info) of this construct
   ci   = compiler info
   local= flag, 1=allocate a new local symbol stack for the duration of this call
*/
void c_fn(SEXP e, int *lp, char *res, SEXP *resy, struct compinfo *ci, int local) {
  const char *fn;
  const char *arop = 0;
  char op[16][64];
  SEXP opy[16];
  int opt[16];
  int ops = 0;
  int symarg0 = 0;    /* is the first argument just symbolic (e.g. in assignments, for, ...)? */
  int new_sym = 0;    /* LHS: is new symbol (needs to be declared)? */
  SEXP sCmd;
  int pct = 0;
  int ftype = 0, fid;

  /* sanity checks */

  if (e == R_NilValue) return;
  if (TYPEOF(e) != LANGSXP)
    error("expecting LANGSXP, found %d", TYPEOF(e));
  sCmd=CAR(e);
  if (TYPEOF(sCmd) != SYMSXP)
    error("expecting symbol as function");
  e=CDR(e);
  fn = CHAR(PRINTNAME(sCmd));

  /* allocate local symbol tabel if necessary */

  if (local) ci->lstc = new_stchain(new_symtab(), ci->lstc);
  /* Rprintf("local symtab=%x (&=%x)\n", (int)ci->lstc->symtab, (int)(&(ci->lstc->symtab))); */

  /* check whether we know this function */

  fid = fn_id(fn);
  ci->usedfn[fid] = 1;
  if (fid>=0) ftype = int_funs[fid].type;


  if (ftype == FNT_COMPOUND) { /* `{` */
    SEXP c = e;
    if (TYPEOF(c) != LISTSXP)
      error("no parameters (%s)", fn);
    if (c != R_NilValue) {
      if (CDR(c) != R_NilValue) {
	while (CDR(c) != R_NilValue) {
	  c_fn(CAR(c), lp, 0, 0, ci, 0); /* discard all results */
	  c = CDR(c);
	}
      }
      /* eval the final result */
      c_fn(CAR(c), lp, res, resy, ci, 0);
      if (pct) UNPROTECT(pct);
      if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
      return;
    }
    /* {} - empty compound -> NULL value */
    if (res) {
      strcpy(res, "R_NilValue");
      resy[0]=R_NilValue;
    }
    if (pct) UNPROTECT(pct);
    if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
    return;
  }

  /* assignment and for need to treat their firs parameter differently */
  if (ftype == FNT_ASSIGN || ftype == FNT_FOR)
    symarg0 = 1;
  
  /* ---- get and evaluate closure parameters ---- */

  if (TYPEOF(e) == LISTSXP) { /* if we have any parameters at all ... */
    SEXP c = e;
    while (c != R_NilValue) {
      SEXP p = CAR(c);
      opt[ops] = TYPEOF(p);
      opy[ops] = 0;
      if (TYPEOF(p) == SYMSXP) {
	const char *sn = CHAR(PRINTNAME(p));

	if (symarg0 && ops==0) { /* first parameter is assignment target */
	  if (!(opy[ops]=find_sym(ci->lstc->symtab, p))) { /* do NOT cascade down */
	    Rprintf("new local symbol %s <%x>\n", sn, (int)p);
	    new_sym = 1;
	    opy[ops]=add_sym(ci->lstc->symtab, p, new_si(R_NilValue, STY_NEW)); /* we should fix this by using info from op[1] */
	  }
	} else {
	  if (!(opy[ops]=find_sym_rec(ci->lstc, p)) && !(opy[ops]=find_sym(ci->gltab, p))) {
	    SEXP v = findVar(p, R_GlobalEnv);
	    if (v == R_UnboundValue) {
	      Rprintf("unbound symbol: %s <%x>, cannot optimize!\n", sn, (int)p);
	      opy[ops]=add_sym(ci->gltab, p, new_si(v, STY_UNKN|SFL_UNB));
	    } else {
	      PROTECT(v);
	      if (TYPEOF(v) == PROMSXP) {
		Rprintf("%s is a promise, evaluating\n", sn);
		v = eval(v, R_GlobalEnv);
		UNPROTECT(1);
		PROTECT(v);
	      }
	      {
		int st = TYPEOF(v);
		int fl = (ATTRIB(v)==R_NilValue)?0:SFL_ATT;
		switch (st) {
		case INTSXP:
		  fl |= STY_INT;
		  if (LENGTH(v)==1) fl |= SFL_SCAL;
		  break;
		case REALSXP:
		  fl |= STY_REAL;
		  if (LENGTH(v)==1) fl |= SFL_SCAL;
		  break;
		case STRSXP:
		  fl |= STY_CHAR;
		  if (LENGTH(v)==1) fl |= SFL_SCAL;
		  break;
		default:
		  fl |= STY_OTHER;
		}
		fl |= SFL_CONST; /* we're taking all global symbols as const for now */
		opy[ops]=add_sym(ci->gltab, p, new_si(v, fl));
		Rprintf("global symbol %s [%d, %x] is assumed to be constant\n", sn, st, fl);
		UNPROTECT(1);
	      }
	    }
	  }
	}
	sprintf(op[ops], "_sym_%s", sn);
	{ char *c = op[ops]; while (*c) { if (*c=='.') *c='_'; c++; }; }
      } else if (TYPEOF(p) == REALSXP) {
	sprintf(op[ops], "ScalarReal(%f)", REAL(p)[0]);
	opy[ops]=new_si(p, SFL_SCAL|SFL_CONST|STY_REAL);
      } else if (TYPEOF(p) == INTSXP) {
	sprintf(op[ops], "ScalarInteger(%d)", INTEGER(p)[0]);
	opy[ops]=new_si(p, SFL_SCAL|SFL_CONST|STY_INT);
      } else if (TYPEOF(p) == STRSXP) {
	sprintf(op[ops], "mkString(\"%s\")", CHAR(STRING_ELT(p, 0)));
	opy[ops]=new_si(STRING_ELT(p, 0), SFL_SCAL|SFL_CONST|STY_CHAR);
      } else if (TYPEOF(p) == LGLSXP) {
	sprintf(op[ops], "ScalarLogical(%d)", INTEGER(p)[0]);
	opy[ops]=new_si(p, SFL_SCAL|SFL_CONST|STY_INT);
      } else if (p == R_NilValue) {
	strcpy(op[ops], "R_NilValue");
	opy[ops]=new_si(p, SFL_CONST|STY_OTHER);
      } else if (TYPEOF(p) == LANGSXP) {
	/* special case: for loop; third parameter is the code, so we're here */
	if (symarg0 && ops==2 && ftype == FNT_FOR) {
	  int ip = ++lp[0];
	  comp_out(ci, "SEXP _is%d = %s;\n", ip, op[1]);
	  if (opy[1]) ADD_SSFLAG(opy[1], SFL_USED);
	  if (res)
	    comp_out(ci, "SEXP _iv%d = R_NilValue;\n", ip);
	  comp_out(ci, "SEXP %s;\n", op[0]);
	  comp_out(ci, "int _ii%d = 0, _il%d = ", ip, ip);
	  if (SSFLAG(opy[1])&SFL_CONST)
	    comp_out(ci, "%d", LENGTH(CAR(opy[1])));
	  else
	    comp_out(ci, "LENGTH(_is%d)", ip);
	  comp_out(ci, ";\nwhile (_ii%d < _il%d) {\n", ip, ip);
	  comp_out(ci, "  if (_ii%d==0)\n    %s = ScalarInteger(INTEGER(_is%d)[0]);\n  else\n    INTEGER(%s)[0] = INTEGER(_is%d)[_ii%d];", ip, op[0], ip, op[0], ip, ip);
	  comp_out(ci, "\n  ");
	  /* evaluate the inside of the for loop */
	  c_fn(p, lp, res, resy, ci, 0);
	  if (res) 
	    comp_out(ci, "  _iv%d = %s;\n", ip, res);
	  comp_out(ci, "  _ii%d++;\n}\n", ip);
	  if (res) {
	    sprintf(res, "_iv%d", ip);
	    if (resy[0]) ADD_SSFLAG(resy[0], SFL_USED);
	  }
	  if (pct) UNPROTECT(pct);
	  if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
	  return;
	}

	/* recursion - the parameter is anothe rlanguage construct that we need to get first */
	c_fn(p, lp, op[ops], &opy[ops], ci, 0);
      } else error("invalid parameter type (%s at %d is %d)", fn, ops, TYPEOF(c));

#if 0
      if (opy[ops])
	Rprintf("/* op[%d]: ty=%02x, fl=%04x */ ", ops, SSTYPE(opy[ops]), SSFLAG(opy[ops]));
      else
	Rprintf("/* op[%d] (?) */ ", ops);
#endif
      if (opy[ops]) { PROTECT(opy[ops]); pct++; }
      ops++;
      c = CDR(c);
    }
  }

  /* CONSTANT FOLDING */
  /* --- check whether the parameters re all constant. if so, evaluate the expression ---*/

  if (ftype != FNT_ASSIGN) { /* anything BUT assignments - see if all pars are constant */
    int all_const = 1;
    int j = 0;
    Rprintf("/* %s: ",fn);
    while (j < ops) {
      if (opy[j])
	Rprintf("(%x,%x) ", SSFLAG(opy[j]), SSTYPE(opy[j]));
      else 
	Rprintf("<null> ");
      if (!opy[j] || opy[j]==R_NilValue || ((SSFLAG(opy[j])&SFL_CONST)==0))
	{ all_const=0; break; };
      j++;
    }
    Rprintf(" */\n");
    if (all_const && FNT_IS_PURE(ftype)) {
      if (!res) { /* no result needed and the fn is pure, hence we can get out right away */
	if (pct) UNPROTECT(pct);
	if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
	return;	
      } else {
	SEXP ex = LCONS(sCmd, R_NilValue), er;
	SEXP ap = ex;
	PROTECT(ex);
	{
	  j = 0;
	  while (j < ops) {
	    SEXP nv = CONS(SVAL(opy[j]), R_NilValue);
	    SETCDR(ap, nv);
	    ap = nv;
	    j++;
	  }
	}
	er = eval(ex, R_GlobalEnv);
	UNPROTECT(1);
	if (classify_val(ci, er, res, resy)) {
#ifdef CHATTY
	  Rprintf("/* "); print_exp(ex); Rprintf(" = "); print_val(er); Rprintf(" = %s */", res);
#endif
	  if (pct) UNPROTECT(pct);
	  if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
	  return;
	}
      }
    }
  }
  
  /* special case : ARITHOPs */

  if (ftype == FNT_ARITHOP) {
    if (!strcmp(fn, "+")) { fn=(ops>1)?"_ar2":"_ar1"; arop="PLUSOP"; }
    if (!strcmp(fn, "-")) { fn=(ops>1)?"_ar2":"_ar1"; arop="MINUSOP"; }
    if (!strcmp(fn, "*")) { fn="_ar2"; arop="TIMESOP"; }
    if (!strcmp(fn, "/")) { fn="_ar2"; arop="DIVOP"; }
    if (!strcmp(fn, "^")) { fn="_ar2"; arop="POWOP"; }
    if (!strcmp(fn, "%%")) { fn="_ar2"; arop="MODOP"; }
    /* if (!strcmp(fn, "^")) { fn="_ar"; arop="IDIVOP"; }*/
  }

  /* special case: COLLECT `(` */

  if (ftype == FNT_COLLECT) {
    if (ops != 1) error("invalid number of ops (%d) to `(`, should be one", ops);
    if (res) {
      strcpy(res, op[0]);
      resy[0] = opy[0];
    }
    if (pct) UNPROTECT(pct);
    if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
    return;
  }
  
  /* special case: ASSIGN */

  if (ftype == FNT_ASSIGN) {
    if (opy[1]) { /* if we have info on the contents, we can propagate it */
      int fl = SSFLAG(opy[1]);
      Rprintf("/* %s = <%x,%x> */", op[0], fl, SSTYPE(opy[1]));
      ADD_SSFLAG(opy[1], SFL_USED);
      if (fl&SFL_CONST) {
	ADD_SSFLAG(opy[0], SFL_CONST);
	SET_SVAL(opy[0], duplicate(SVAL(opy[1])));
      } else {
	/* fix type if it is non-const yet available */
      }
    } else {
      Rprintf("/* %s = *type-unknown* %s */", op[0], op[1]);
      SET_SSFLAG(opy[0], 0);
    }
    comp_out(ci, "%s%s = %s;\n", new_sym?"SEXP ":"", op[0], op[1]);
    if (res) {
      strcpy(res, op[0]);
      resy[0]=opy[0];
    }
    if (pct) UNPROTECT(pct);
    if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
    return;
  }

  /* --- create a local variable with the result --- */

  if (res) {
    lp[0]++;
    comp_out(ci, "SEXP _l%d = ", *lp);
  }

  if (!FNT_CLASS(ftype))
    Rprintf("Warning: native function %s unimplemented!", fn);

  comp_out(ci, "RC_%s(", fn);
  {
    int i = 0;
    if (arop) comp_out(ci, "%s, ", arop);
    while (i < ops) {
      comp_out(ci, "%s", op[i]);
      if (i < ops-1) comp_out(ci, ", ");
      /* mark as used */
      if (opy[i]) ADD_SSFLAG(opy[i], SFL_USED);
      i++;
    }
    /* comp_out(ci, "); PROTECT(_l%d);\n", *lp); */
    comp_out(ci, ");\n");
  }
  if (res)
    sprintf(res, "_l%d", *lp);
  if (pct) UNPROTECT(pct);
  if (local) ci->lstc = free_stchain(ci->lstc); /* pop local table */
}


/* --- R compilation driver --- */

SEXP c(SEXP e, SEXP fn) {
  int lp = 0;
  char buf[64];
  struct compinfo ci;
  SEXP rsym = 0;
  SEXP argnames = R_NilValue;
  SEXP argtail = 0;
  int argncount = 0;
  *buf=0;
  initCI(&ci);
  if (TYPEOF(e) == EXPRSXP) e = VECTOR_ELT(e, 0);
  c_fn(e, &lp, buf, &rsym, &ci, 1);
  /* return value has to be marked as used */
  if (rsym) ADD_SSFLAG(rsym, SFL_USED);
  {
    SEXP gt = CAR(ci.gltab);
#ifdef CALL_SIG
    int firstpar = 1;
#endif
    struct lbuf *o_decl = pre_str(ci.outbuf->head,"\n/* function declaration */\nSEXP ");
    struct lbuf *o_init = pre_str(o_decl, "\nvoid R_init_");
    struct lbuf *o_consts = pre_str(o_init,"#include \"basis.c\"\n\n/* global constants */\n");
    insert_str(o_decl, CHAR(STRING_ELT(fn, 0)));
    insert_str(o_decl,"(");
#ifndef CALL_SIG
    insert_str(o_decl,"SEXP _op, SEXP _rho) {\n");
#endif
    insert_str(o_init, CHAR(STRING_ELT(fn, 0)));
    insert_str(o_init,"() {\n");

    /* collect all used math functions */
    {
      char ln[64];
      int i = 0;
      while (i < INT_FNS) {
	if (ci.usedfn[i] && int_funs[i].type==FNT_MATH1) {
	  sprintf(ln, "MAKE_MATH1_FN(%s)\n", int_funs[i].fn);
	  o_consts=insert_str(o_consts, ln);
	}
	i++;
      }
    }

    /* collect all unbound globals that will turn into arguments */
    while (TAG(gt) != R_NilValue) {
      char b[128];
      char ini[512], ln[512];
      SEXP sym = TAG(gt);
      int fl = SSFLAG(CAR(gt));
      int ty = SSTYPE(CAR(gt));
      strcpy(b, CHAR(PRINTNAME(sym)));
      *ini=0;
      if (fl&SFL_CONST) {
	SEXP cres = 0;
	classify_val(&ci, CAR(CAR(gt)), ini, &cres);
	if (cres) ADD_SSFLAG(cres, SFL_USED);
      }
      { char *c=b; while(*c) { if (*c=='.') *c='_'; c++; } }
      if (fl&SFL_CONST) {
	sprintf(ln, "SEXP _sym_%s; /* %02x, %02x */\n", b, fl, ty);
	o_consts = insert_str(o_consts, ln);
	if (*ini) {
	  sprintf(ln, "  _sym_%s = %s; /* %02x, %02x */\n", b, ini, fl, ty);
	  o_init = insert_str(o_init, ln);
	}
      } else {
#ifdef CALL_SIG
	sprintf(ln, "%sSEXP _sym_%s", firstpar?"":", ", b);
	o_decl = insert_str(o_decl, ln);
	firstpar=0;
#else
	/* sprintf(ln, "SEXP _sym_%s = CAR(_op); _op = CDR(_op);\n", b);*/
	sprintf(ln, "SEXP NC_GETOP(_sym_%s, _op); _op = CDR(_op);\n", b);
	o_decl = insert_str(o_decl, ln);	
#endif
	if (argtail) {
	  SEXP nt = CONS(mkChar(b), R_NilValue);
	  SETCDR(argtail, nt);
	  argtail = nt;
	} else {
	  PROTECT(argnames = CONS(mkChar(b), R_NilValue));
	  argtail = argnames;
	}
	argncount++;
      }
      gt = CDR(gt);
    }

    /* collect all constants */
    {
      SEXP ct = CAR(ci.ctab);
      while (TAG(ct) != R_NilValue) {
	int fl = SSFLAG(CAR(ct));
	if (fl&SFL_USED) {
	  char ll[64];
	  const char *cn = CHAR(PRINTNAME(TAG(ct)));
	  SEXP v = CAR(CAR(ct));
	  R_PreserveObject(v);
	  Rprintf("%s = %p (%d)\n", cn, v, TYPEOF(v));
	  sprintf(ll, "SEXP %s = (SEXP) %p;\n", cn, (void*) v);
	  o_consts = insert_str(o_consts, ll);
	}
	ct = CDR(ct);
      }
    }

#ifdef CALL_SIG
    o_decl = insert_str(o_decl, ") {\n");
#endif
    o_init = insert_str(o_init, "}\n");
    
    /* if (lp)
       comp_out(&ci, "UNPROTECT(%d);\n", lp); */
    /* if (argncount)
       comp_out(&ci, "UNPROTECT(%d);\n", argncount); */
    comp_out(&ci, "return %s;\n}\n", buf);

    {
      const char *cfn = CHAR(STRING_ELT(fn, 0));
      comp_out(&ci, "\n#ifdef NATFSXP\nSEXP get_%s() { return R_MakeNativeFunction(%s); }\n#else\nSEXP get_%s() { SEXP f=allocSExp(CLOSXP); SEXP fo=CONS(R_MissingArg,R_NilValue); SET_TAG(fo,R_DotsSymbol); SET_CLOENV(f,R_GlobalEnv);SET_FORMALS(f, fo);SET_BODY(f,LCONS(install(\".Call\"),CONS(mkString(\"%s\"),CONS(R_DotsSymbol, R_NilValue)))); return f; }\n#endif\n", cfn, cfn, cfn, cfn);
    }

    {
      char *res = concat_str(o_consts);
      SEXP ans = mkString(res);
      SEXP an = allocVector(STRSXP, argncount);
      PROTECT(an);
      if (argncount) {
	int i = 0;
	argtail = argnames;
	while (i < argncount) {
	  SET_STRING_ELT(an, i, CAR(argtail));
	  argtail = CDR(argtail);
	  i++;
	}
      }
      setAttrib(ans, install("argnames"), an);
      UNPROTECT(argncount?2:1);
      free(res);
      freeCI(&ci);
      return ans;
    }
  }
}
